//---------------------------------------------------------------------------
/*
  Working with RVF and RTF files containing shared images.
  This demo stores them in the special subdirectory, but you can store them
  in a database, etc.
  How to save pictures as references in RVF:
    rvfoSavePicturesBody is EXCLUDED from RichViewEdit1.RVFOptions.
  How to save pictures as references in RTF:
    see RichViewEdit1SaveItemToFile.
  How to load pictures when reading such RVF files:
    see RichViewEdit1RVFPictureNeeded.
  How to load pictures when reading such RTF files:
    see RichViewEdit1ImportPicture.

  Paths to image file names are stored in rvespImageFileName properties.
  Normally, full paths to images are stored there.
  In this demo, path is stored relative to the application directory.
  Saved RTF and HTML files will be opened normally in other applications
  only if they are stored in the application directory.
*/
//---------------------------------------------------------------------------
/*
  Version 3.1
  - Using RVGraphicHandler
  Version 3.0
  Changes in v3.0:
  - compatibility with TRichView 14 (new parameters in OnRVFPictureNeeded
  - table and table cell background images are supported (in RVF)
  Changes in v2.0:
  - RTF support;
  - HTML saving;
  - storing file names in rvespImageFileName property instead of item names;
  - storing paths relative to the application path instead of storing just
    file names (allows external applications to open RTF and HTML files
    correctly, if they are saved in the application path)
*/
//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <stdlib.h>
#pragma hdrstop

#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma link "RVEdit"
#pragma link "RichView"
#pragma link "RVScroll"
#pragma link "RVStyle"
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
  Randomize();
  OpenDialog2->InitialDir = ExtractFilePath(Application->ExeName);
  SaveDialog1->InitialDir = ExtractFilePath(Application->ExeName);
  RichViewEdit1->LoadRVF(ExtractFilePath(Application->ExeName)+"demo.rvf");
  RichViewEdit1->Format();
}
//---------------------------------------------------------------------------
// Returns graphic loaded from FileName.
// If loading fails, returns RVStyle1.InvalidPicture
TGraphic* TForm1::LoadGraphic(const AnsiString FileName)
{
 TGraphic* gr = RVGraphicHandler->LoadFromFile(FileName);
 if (!gr)
 {
   gr = RVGraphicHandler->CreateGraphic(
     (TGraphicClass)RVStyle1->InvalidPicture->Graphic->ClassType());
   gr->Assign(RVStyle1->InvalidPicture->Graphic);
 }
  return gr;
}
//---------------------------------------------------------------------------
// RichViewEdit1->OnRVFPictureNeeded
// This event occurs when reading RVF files.
// This event load this image from the Images subdirectory.
void __fastcall TForm1::RichViewEdit1RVFPictureNeeded(
      TCustomRichView *Sender, const AnsiString ItemName,
      TRVNonTextItemInfo *Item, int Index1, int Index2, TGraphic *&gr)
{
  AnsiString FileName;
  if (Item->InheritsFrom(__classid(TRVGraphicItemInfo)))
    FileName = ((TRVGraphicItemInfo*)Item)->ImageFileName;
  else if (Item->InheritsFrom(__classid(TRVTableItemInfo)))
    if (Index1>=0 && Index2>=0)
      FileName = ((TRVTableItemInfo*)Item)->Cells[Index1][Index2]->BackgroundImageFileName;
    else
      FileName = ((TRVTableItemInfo*)Item)->BackgroundImageFileName;
  else
    FileName = "";
  if (FileName.IsEmpty())
    return;

  FileName = ExtractFilePath(Application->ExeName)+FileName; // path is relative to the application path
  gr = LoadGraphic(FileName);
}
//---------------------------------------------------------------------------
// RichViewEdit1.OnImportPicture
// This event occurs when loading RTF files with external pictures
void __fastcall TForm1::RichViewEdit1ImportPicture(TCustomRichView *Sender,
      const AnsiString Location, int Width, int Height, TGraphic *&Graphic)
{
  AnsiString FileName =
    ExtractRelativePath(ExtractFilePath(Application->ExeName), Location);
  Graphic = LoadGraphic(FileName);
}
//---------------------------------------------------------------------------
// RichViewEdit1.OnSaveItemToFile
// Modifying how pictures are saved in RTF: saving as external pictures
void __fastcall TForm1::RichViewEdit1SaveItemToFile(
      TCustomRichView *Sender, const AnsiString Path,
      TCustomRVData *RVData, int ItemNo, TRVSaveFormat SaveFormat,
      bool Unicode, TRVRawByteString &OutStr, bool &DoDefault)
{
 if (SaveFormat==rvsfRTF &&
    (RVData->GetItemStyle(ItemNo)==rvsPicture ||
     RVData->GetItemStyle(ItemNo)==rvsHotPicture))
 {
    AnsiString FileName;
    RVData->GetItemExtraStrProperty(ItemNo, rvespImageFileName, FileName);
    OutStr = Format("{\\field{\\*\\fldinst INCLUDEPICTURE \"%s\" \\\\d }}",
      ARRAYOFCONST((
        RVMakeRTFFileNameStr(FileName,
          RichViewEdit1->Style->DefCodePage,
          RichViewEdit1->RTFOptions.Contains(rvrtfDuplicateUnicode)).c_str())));
    DoDefault = false;
 }
}
//---------------------------------------------------------------------------
// Inserting image.
// If this image is not from the Images subdirectory, copying it there
// (under the unique file name)
// Image file name is written in rvespImageFileName
//  (relative to the application path)
void __fastcall TForm1::Button1Click(TObject *Sender)
{
  if (OpenDialog1->Execute())
  {
    TGraphic* gr = RVGraphicHandler->LoadFromFile(OpenDialog1->FileName);
    if (gr)
    {
      AnsiString ImageName = ExtractRelativePath(ExtractFilePath(Application->ExeName),
        CopyImageToTheImagesDir(OpenDialog1->FileName, NULL));
      RichViewEdit1->TopLevelEditor->BeginUndoGroup(rvutInsert);
      RichViewEdit1->TopLevelEditor->SetUndoGroupMode(true);
      try
      {
        if (RichViewEdit1->InsertPicture("", gr, rvvaBaseline))
          RichViewEdit1->SetCurrentItemExtraStrProperty(rvespImageFileName, ImageName, true);
      }
      __finally
      {
        RichViewEdit1->TopLevelEditor->SetUndoGroupMode(false);
      }
    }
    else
      Application->MessageBox("Image loading error", "Error", 0);
  }
}
//---------------------------------------------------------------------------
// Copying the file ImageFileName to the images subdirectory (if gr==NULL)
// or saving gr in the images subdirectory.
// Assigning an unique file name.
// Both ImageFileName and returned value are fully qualified paths.
AnsiString TForm1::CopyImageToTheImagesDir(AnsiString ImageFileName, TGraphic* gr)
{
  AnsiString ImagesDir, NewImageFileName, ImageExt;
  int RandomValue;

  ImageFileName = AnsiLowerCase(ImageFileName);
  ImagesDir = AnsiLowerCase(ExtractFilePath(Application->ExeName)+"Images\\");
  if (ImageFileName.Pos(ImagesDir)!=1)
  {
    NewImageFileName = ImagesDir+ExtractFileName(ImageFileName);
    if (FileExists(NewImageFileName))
    {
      ImageExt = ExtractFileExt(NewImageFileName);
      NewImageFileName = NewImageFileName.SubString(1, NewImageFileName.Length()-ImageExt.Length());
      RandomValue = random(MaxInt);
      while (FileExists(NewImageFileName+IntToStr(RandomValue)+ImageExt))
        RandomValue++;
      NewImageFileName = NewImageFileName+IntToStr(RandomValue)+ImageExt;
    }
    if (!gr)
      CopyFile(ImageFileName.c_str(), NewImageFileName.c_str(), false);
    else
      gr->SaveToFile(NewImageFileName);
    return NewImageFileName;
  }
  else
    return ImageFileName;
}
//---------------------------------------------------------------------------
// Saving all images that not in the images directory
// Such images can appear when loading or pasting files with images
void TForm1::SaveAllUnknownImages(TCustomRVData* RVData)
{
  AnsiString ImageFileName, Ext;
  TGraphic* gr;
  for (int i=0; i<RVData->ItemCount; i++)
    switch (RVData->GetItemStyle(i))
    {
      case rvsPicture:
      case rvsHotPicture:
      {
        // pictures
        RVData->GetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
        ImageFileName = ExtractFilePath(Application->ExeName)+ImageFileName;
        if (!FileExists(ImageFileName))
        {
          TRVVAlign VAlign;
          AnsiString s;
          TRVTag Tag;
          RVData->GetPictureInfo(i, s, gr, VAlign, Tag);
          Ext = GraphicExtension((TGraphicClass)gr->ClassType());
          ImageFileName = ExtractRelativePath(ExtractFilePath(Application->ExeName),
            CopyImageToTheImagesDir("Image."+Ext, gr));
          RVData->SetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
        }
        break;
      }
      case rvsTable:
      {
        TRVTableItemInfo* table = (TRVTableItemInfo*)(RVData->GetItem(i));
        if (table->BackgroundImage)
        {
          // table background image
          ImageFileName = ExtractFilePath(Application->ExeName)+
            table->BackgroundImageFileName;
          if (!FileExists(ImageFileName))
          {
            gr = table->BackgroundImage;
            Ext = GraphicExtension((TGraphicClass)gr->ClassType());
            ImageFileName = ExtractRelativePath(ExtractFilePath(Application->ExeName),
              CopyImageToTheImagesDir("Image."+Ext, gr));
            table->BackgroundImageFileName = ImageFileName;
          }
        }
        for (int r=0; r<table->Rows->Count; r++)
          for (int c=0; c<table->Rows->Items[r]->Count; c++)
            if (table->Cells[r][c])
            {
              if (table->Cells[r][c]->BackgroundImage)
              {
                // cell background image
                ImageFileName = ExtractFilePath(Application->ExeName)+
                  table->Cells[r][c]->BackgroundImageFileName;
                if (!FileExists(ImageFileName))
                {
                  gr = table->Cells[r][c]->BackgroundImage;
                  Ext = GraphicExtension((TGraphicClass)gr->ClassType());
                  ImageFileName = ExtractRelativePath(ExtractFilePath(Application->ExeName),
                    CopyImageToTheImagesDir("Image."+Ext, gr));
                   table->Cells[r][c]->BackgroundImageFileName = ImageFileName;
                }
              }
              SaveAllUnknownImages(table->Cells[r][c]->GetRVData());
            }
        break;
      }
    }
}
//---------------------------------------------------------------------------
// After loading from RTF, full file names are assigned to rvespImageFileName
// properties. This procedure is called to convert them to relative paths.
void TForm1::ConvertAllPathsToRelativePaths(TCustomRVData* RVData)
{
  for (int i=0; i<RVData->ItemCount; i++)
    switch (RVData->GetItemStyle(i))
    {
      case rvsPicture:
      case rvsHotPicture:
      {
        AnsiString ImageFileName;
        RVData->GetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
        ImageFileName = ExtractRelativePath(ExtractFilePath(Application->ExeName), ImageFileName);
        RVData->SetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
        break;
      }
      case rvsTable:
      {
        TRVTableItemInfo* table = (TRVTableItemInfo*)(RVData->GetItem(i));
        for (int r=0; r<table->Rows->Count; r++)
          for (int c=0; c<table->Rows->Items[r]->Count; c++)
            if (table->Cells[r][c])
              ConvertAllPathsToRelativePaths(table->Cells[r][c]->GetRVData());
        break;
      }
    }
}
//---------------------------------------------------------------------------
// Before copying to the clipboard
void __fastcall TForm1::RichViewEdit1Copy(TObject *Sender)
{
  SaveAllUnknownImages(RichViewEdit1->RVData);
}
//---------------------------------------------------------------------------
// Loading doc
void __fastcall TForm1::Button3Click(TObject *Sender)
{
  if (OpenDialog2->Execute())
  {
    RichViewEdit1->Clear();
    bool r = false;
    switch (OpenDialog2->FilterIndex)
    {
      case 1:
        r = RichViewEdit1->LoadRVF(OpenDialog2->FileName);
        break;
      case 2:
        r = RichViewEdit1->LoadRTF(OpenDialog2->FileName);
        ConvertAllPathsToRelativePaths(RichViewEdit1->RVData);
        break;
    }
    RichViewEdit1->Format();
    if (!r)
      Application->MessageBox("Document loading error", "Error", 0);
  }
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Button2Click(TObject *Sender)
{
  if (SaveDialog1->Execute())
  {
    SaveAllUnknownImages(RichViewEdit1->RVData);
    bool r = false;
    switch (SaveDialog1->FilterIndex)
    {
      case 1:
        r = RichViewEdit1->SaveRVF(SaveDialog1->FileName, false);
        break;
      case 2:
        r = RichViewEdit1->SaveRTF(SaveDialog1->FileName, false);
        break;
      case 3:
        TRVSaveOptions SaveOptions = TRVSaveOptions();
        SaveOptions << rvsoUseCheckpointsNames;
        SaveOptions << rvsoUseItemImageFileNames;
        r = RichViewEdit1->SaveHTMLEx(SaveDialog1->FileName, "Shared Image Demo",
        "img", "", "", "", SaveOptions);
    }
    if (!r)
      Application->MessageBox("Document saving error", "Error", 0);
  }
}
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------


//---------------------------------------------------------------------------

