/*==============================================================================
  RichEditor Demo with StyleTemplates
  Menu items disabling/enabling is not implemented here.

  This demo shows all main operations with StyleTemplates:
  - loading default style templates in File | New;
  - displaying information about formatting at the position of the caret;
  - implementing a style combo-box;
  - clear format command;
  - editing styles.

  For simplification, this demo supports a very limited set of text and paragraph attributes.

  IMPORTANT: If you right click the editor in design time, choose "Settings"
  in the context menu, you'll see that the radiogroup is in state
  "Allow adding styles dynamically + style templates"
==============================================================================*/
#include <vcl.h>
#pragma hdrstop

#include "Unit1.h"
#include "StyleListFrm.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "RichView"
#pragma link "RVEdit"
#pragma link "RVScroll"
#pragma link "RVStyle"
#pragma resource "*.dfm"
TfrmMain *frmMain;
//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
// Parameters for ApplyStyleConversion
#define TEXT_BOLD          1
#define TEXT_ITALIC        2
#define TEXT_UNDERLINE     3
#define TEXT_APPLYFONTNAME 4
#define TEXT_APPLYFONTSIZE 5
// Parameters for ApplyParaStyleConversion
#define PARA_ALIGNMENT     1
void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
  // Filling font names combobox
  cmbFont->Items->Assign(Screen->Fonts);
  New();
}
//---------------------------------------------------------------------------
// current text style was changed
void __fastcall TfrmMain::rveCurTextStyleChanged(TObject *Sender)
{
  IgnoreChanges = true;

  // Changing selection in comboboxes with font names and sizes:
  TFontInfo* fi = rvs->TextStyles->Items[rve->CurTextStyleNo];
  cmbFont->ItemIndex = cmbFont->Items->IndexOf(fi->FontName);
  cmbFontSize->Text = IntToStr(fi->Size);

  // Checking font buttons
  btnBold->Down      = fi->Style.Contains(fsBold);
  btnItalic->Down    = fi->Style.Contains(fsItalic);
  btnUnderline->Down = fi->Style.Contains(fsUnderline);

  // Styles
  SelectInStyleComboBox();
  DescribeFormatting();

  IgnoreChanges = false;
}
//---------------------------------------------------------------------------
// current paragraph style was changed
void __fastcall TfrmMain::rveCurParaStyleChanged(TObject *Sender)
{
  SetAlignmentToUI(rvs->ParaStyles->Items[rve->CurParaStyleNo]->Alignment);
  SelectInStyleComboBox();
  DescribeFormatting();
}
//--------------------------------------------------------------------------
void __fastcall TfrmMain::rveChange(TObject *Sender)
{
  DescribeFormatting();
}
//--------------------------------------------------------------------------
TRVAlignment TfrmMain::GetAlignmentFromUI()
{
  if (btnLeft->Down)
    return rvaLeft;
  else if (btnRight->Down)
    return rvaRight;
  else if (btnCenter->Down)
    return rvaCenter;
  else
    return rvaJustify;
}
//--------------------------------------------------------------------------
void TfrmMain::SetAlignmentToUI(TRVAlignment Alignment)
{
  switch (Alignment)
  {
    case rvaLeft:
      btnLeft->Down = true;
      break;
    case rvaCenter:
      btnCenter->Down = true;
      break;
    case rvaRight:
      btnRight->Down = true;
      break;
    case rvaJustify:
      btnJustify->Down = true;
      break;
  }
}
//---------------------------------------------------------------------------
// applying font name
void __fastcall TfrmMain::cmbFontClick(TObject *Sender)
{
  if (cmbFont->ItemIndex!=-1 && !IgnoreChanges)
  {
    FontName = cmbFont->Items->Strings[cmbFont->ItemIndex];
    rve->ApplyStyleConversion(TEXT_APPLYFONTNAME, true);
  }
  if (Visible)
    rve->SetFocus();
}
//---------------------------------------------------------------------------
// applying font size
void __fastcall TfrmMain::cmbFontSizeClick(TObject *Sender)
{
  if (cmbFontSize->Text!="" && !IgnoreChanges)
  {
    FontSize = StrToIntDef(cmbFontSize->Text, 10);
    rve->ApplyStyleConversion(TEXT_APPLYFONTSIZE, true);
  }
  if (Visible)
    rve->SetFocus();
}
//---------------------------------------------------------------------------
// bold, italic, underline
void __fastcall TfrmMain::FontStyleButtonClick(TObject *Sender)
{
  TSpeedButton*Button = (TSpeedButton*)Sender;
  // constants TEXT_BOLD, TEXT_ITALIC and TEXT_UNDERLINE are
  // assigned to the tags of corresponding buttons
  rve->ApplyStyleConversion(Button->Tag, true);
}
//---------------------------------------------------------------------------
// applying paragraph alignment
void __fastcall TfrmMain::btnApplyPara(TObject *Sender)
{
   rve->ApplyParaStyleConversion(PARA_ALIGNMENT, true);
}
//---------------------------------------------------------------------------
// rve->OnStyleConversion
void __fastcall TfrmMain::rveStyleConversion(TCustomRichViewEdit *Sender,
      int StyleNo, int UserData, bool AppliedToText, int &NewStyleNo)
{
  TFontInfo* FontInfo = new TFontInfo(NULL);
  FontInfo->Assign(rvs->TextStyles->Items[StyleNo]);
  switch (UserData)
  {
    case TEXT_BOLD:
    {
      if (btnBold->Down)
        FontInfo->Style << fsBold;
      else
        FontInfo->Style >> fsBold;
      break;
    }
    case TEXT_ITALIC:
    {
      if (btnItalic->Down)
        FontInfo->Style << fsItalic;
      else
        FontInfo->Style >> fsItalic;
      break;
    }
    case TEXT_UNDERLINE:
    {
      if (btnUnderline->Down)
        FontInfo->Style << fsUnderline;
      else
        FontInfo->Style >> fsUnderline;
      break;
    }
    case TEXT_APPLYFONTNAME:
      FontInfo->FontName = FontName;
      break;
    case TEXT_APPLYFONTSIZE:
      FontInfo->Size     = FontSize;
      break;
      // add your code here....
  }
  NewStyleNo = rvs->FindTextStyle(FontInfo);
  delete FontInfo;
}
//---------------------------------------------------------------------------
// rve->OnParaStyleConversion
void __fastcall TfrmMain::rveParaStyleConversion(TCustomRichViewEdit *Sender,
	int StyleNo, int UserData, bool AppliedToText, int &NewStyleNo)
{
  TParaInfo* ParaInfo = new TParaInfo(NULL);
  ParaInfo->Assign(rvs->ParaStyles->Items[StyleNo]);
  switch (UserData)
  {
    case PARA_ALIGNMENT:
      ParaInfo->Alignment = GetAlignmentFromUI();
      break;
    // add your code here....
  }
  NewStyleNo = rvs->FindParaStyle(ParaInfo);
  delete ParaInfo;
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::cmbFontSizeKeyPress(TObject *Sender, char &Key)
{
  if (Key==VK_RETURN)
  {
    Key = 0;
    cmbFontSizeClick(NULL);
  }
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::cmbFontSizeExit(TObject *Sender)
{
  cmbFontSizeClick(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitUndoClick(TObject *Sender)
{
  rve->Undo();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitRedoClick(TObject *Sender)
{
  rve->Redo();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitCutClick(TObject *Sender)
{
  rve->CutDef();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitCopyClick(TObject *Sender)
{
  rve->CopyDef();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitPasteClick(TObject *Sender)
{
  rve->Paste();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitDeleteClick(TObject *Sender)
{
  rve->DeleteSelection();
}
//---------------------------------------------------------------------------
bool TfrmMain::SaveIfNeeded()
{
  if (rve->Modified)
    switch (Application->MessageBox(L"Save file now?", L"File was modified",
                                MB_ICONQUESTION | MB_YESNOCANCEL))
    {
      case IDYES:
        return Save();
      case IDNO:
        return true;
      default: //case IDCANCEL:
        return false;
    }
  else
    return true;
}
//---------------------------------------------------------------------------
bool TfrmMain::Save()
{
  if (FileName=="")
    return SaveAs();
  else
  {
    bool r = rve->SaveRVF(FileName, false);
    if (r)
    {
      rve->Modified = false;
      return true;
    }
    else
    {
      Application->MessageBox(L"Cannot save file", L"Error", MB_OK | MB_ICONSTOP);
      return SaveAs();
    }
  }
}
//---------------------------------------------------------------------------
bool TfrmMain::SaveAs()
{
  if (sd->Execute())
  {
    FileName = sd->FileName;
    bool r = Save();
    if (r)
      Caption = ExtractFileName(FileName) + "- RDemo";
    return r;
  }
  else
    return false;
}
//---------------------------------------------------------------------------
void TfrmMain::Open()
{
  if (!SaveIfNeeded())
    return;
  rve->Modified = false;
  New();
  if (od->Execute())
  {
    FileName = od->FileName;
    bool r = rve->LoadRVF(FileName);
    rve->Format();
    rveCurTextStyleChanged(NULL);
    rveCurParaStyleChanged(NULL);
    Caption = ExtractFileName(FileName) + "- RDemo";
    if (!r)
      Application->MessageBox(L"Error while opening or reading file", L"Error",
        MB_OK | MB_ICONSTOP);
  }
}
//---------------------------------------------------------------------------
UnicodeString GetFullFileName(const UnicodeString FileName)
{
  UnicodeString r = ExtractFilePath(Application->ExeName)+FileName;
  if (FileExists(r))
    return r;
  r = ExtractFilePath(Application->ExeName)+"..\\"+FileName;
  if (FileExists(r))
	return r;
  r = ExtractFilePath(Application->ExeName)+"..\\..\\"+FileName;
  if (FileExists(r))
	return r;
  return "";
}
//---------------------------------------------------------------------------
void TfrmMain::New()
{
  if (!SaveIfNeeded())
    return;
  FileName = "";
  Caption = "Unnamed - RDemo";
  rve->LeftMargin   = 5;
  rve->RightMargin  = 5;
  rve->TopMargin    = 5;
  rve->BottomMargin = 5;
  rve->BackgroundBitmap = NULL;
  rve->BackgroundStyle = bsNoBitmap;
  rve->Clear();
  // loading default style templates
  rvs->StyleTemplates->LoadFromRVST(GetFullFileName("Default.rvst"), rvs->Units);
  // creating default text and paragraph styles
  rvs->TextStyles->Clear();
  rvs->ParaStyles->Clear();
  rvs->ListStyles->Clear();
  rvs->ParaStyles->Add();
  rvs->TextStyles->Add();
  // formatting them according to "Normal" style template
  if (rvs->StyleTemplates->NormalStyleTemplate)
  {
    rvs->StyleTemplates->NormalStyleTemplate->ApplyToParaStyle(rvs->ParaStyles->Items[0]);
    ((TRVStyleTemplate*)NULL)->ApplyToTextStyle(rvs->TextStyles->Items[0],
      rvs->StyleTemplates->NormalStyleTemplate);
    rvs->TextStyles->Items[0]->ParaStyleTemplateId = rvs->StyleTemplates->NormalStyleTemplate->Id;
  }
  rve->Format();
  rveStyleTemplatesChange(rve);  
  rveCurTextStyleChanged(NULL);
  rveCurParaStyleChanged(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitNewClick(TObject *Sender)
{
  New();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitOpenClick(TObject *Sender)
{
  Open();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitSaveClick(TObject *Sender)
{
  Save();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitSaveAsClick(TObject *Sender)
{
  SaveAs();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::mitExitClick(TObject *Sender)
{
  Close();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
  CanClose = SaveIfNeeded();
}
//============================ Style Templates =================================
// rve.OnStyleTemplatesChange: filling cmbStyles combo-box
void __fastcall TfrmMain::rveStyleTemplatesChange(TObject *Sender)
{
  TRVStyleTemplateKinds Kinds = TRVStyleTemplateKinds();
  Kinds << rvstkText;
  Kinds << rvstkPara;
  Kinds << rvstkParaText;
  rvs->StyleTemplates->AssignToStrings(cmbStyles->Items, true,
    Kinds, NULL, false);
}
//---------------------------------------------------------------------------
// Selecting in cmbStyles according to the current text and paragraph style
void TfrmMain::SelectInStyleComboBox()
{
  if (! rve->ItemCount)
    return;
  // selecting text style template, if possible
  TRVStyleTemplateId Id = rvs->TextStyles->Items[rve->CurTextStyleNo]->StyleTemplateId;
  if (Id>0)
  {
    cmbStyles->ItemIndex = cmbStyles->Items->IndexOfObject((TObject*)Id);
    return;
  }
  // otherwise, selecting paragraph style template, if possible
  Id = rvs->ParaStyles->Items[rve->CurParaStyleNo]->StyleTemplateId;
  if (Id>0)
  {
    cmbStyles->ItemIndex = cmbStyles->Items->IndexOfObject((TObject*)Id);
    return;
  }
  cmbStyles->ItemIndex = -1;
}
//---------------------------------------------------------------------------
// Applying the style template selected in cmbStyles
void __fastcall TfrmMain::cmbStylesClick(TObject *Sender)
{
  if (cmbStyles->ItemIndex<0)
    return;
  rve->ApplyStyleTemplate(cmbStyles->ItemIndex, true);
  rve->SetFocus();
}
//---------------------------------------------------------------------------
// Clear format
void __fastcall TfrmMain::btnClearFormatClick(TObject *Sender)
{
  rve->ApplyStyleTemplate(-1, true);
}
//---------------------------------------------------------------------------
UnicodeString DescribeFontStyle(TCustomRVFontInfo* TextInfo,
  TRVFontInfoProperties& TextProps, TRVFontInfoProperty Prop, TFontStyle Style,
  const UnicodeString Yes, const UnicodeString No)
{
  UnicodeString r;
  if (TextProps.Contains(Prop))
  {
    if (TextInfo->Style.Contains(Style))
      r = Yes;
    else
      r = No;
    TextProps >> Prop;
    r +=  "; ";
  }
  return r;
}
// Returns a string describing TextInfo. Only properties included in TextProps
// are described.
// (this is only a demo function; it describes only font name, size, bold,
// italic, and underline attributes)
UnicodeString DescribeTextProps(TCustomRVFontInfo* TextInfo,
  TRVFontInfoProperties TextProps)
{
  UnicodeString r =
    DescribeFontStyle(TextInfo, TextProps, rvfiBold, fsBold, "bold", "not bold")+
    DescribeFontStyle(TextInfo, TextProps, rvfiItalic, fsItalic, "italic", "not italic")+
    DescribeFontStyle(TextInfo, TextProps, rvfiUnderline, fsUnderline, "underlined", "not underlined");
  if (TextProps.Contains(rvfiFontName))
  {
    r += "font name "+TextInfo->FontName+"; ";
    TextProps >> rvfiFontName;
  }
  if (TextProps.Contains(rvfiSize))
  {
    r += "font size "+IntToStr(TextInfo->Size)+"; ";
    TextProps >> rvfiSize;
  }
  if (TextProps != TRVFontInfoProperties())
    r += "some other attributes;";
  r = Trim(r);
  if (!r.IsEmpty() && r[r.Length()]==';')
    r.Delete(r.Length(), 1);
  return r;
}
//---------------------------------------------------------------------------
// Returns a string describing ParaInfo. Only properties included in ParaProps
// are described.
// (this is only a demo function; it describes only alignment and heading level)
UnicodeString DescribeParaProps(TCustomRVParaInfo* ParaInfo,
  TRVParaInfoProperties ParaProps)
{
  UnicodeString r;
  if (ParaProps.Contains(rvpiAlignment))
  {
    switch (ParaInfo->Alignment)
    {
      case rvaLeft: r = "left"; break;
      case rvaRight: r = "right"; break;
      case rvaCenter: r = "center"; break;
      default: r = "justify"; break;
    }
    r = "alignment "+r+"; ";
    ParaProps >> rvpiAlignment;
  }
  if (ParaProps.Contains(rvpiOutlineLevel))
  {
    if (ParaInfo->OutlineLevel<=0)
      r += "outline level: body text; ";
    else
      r += "outline level "+IntToStr(ParaInfo->OutlineLevel)+"; ";
    ParaProps >> rvpiOutlineLevel;
  }
  if (ParaProps != TRVParaInfoProperties())
    r += "some other attributes;";
  r = Trim(r);
  if (!r.IsEmpty() && r[r.Length()]==';')
    r.Delete(r.Length(), 1);
  return r;
}
//---------------------------------------------------------------------------
// Returns a string describing paragraph properties of StyleTemplate and
// its parents. Level must be 0.
UnicodeString DescribeStyleTemplatePara(TRVStyleTemplate* StyleTemplate,
  int Level)
{
  if (!StyleTemplate)
    return "none";

  UnicodeString r = "\""+StyleTemplate->Name+"\"\r\n";
  UnicodeString s = DescribeParaProps(StyleTemplate->ParaStyle, StyleTemplate->ValidParaProperties);
  if (!s.IsEmpty())
    r += "("+s+"), ";
  if (Level==0)
    r += "based on ";
  else
    r += "which is based on ";
  if (! StyleTemplate->Parent)
    r += "default paragraph attributes";
  else
    r += DescribeStyleTemplatePara(StyleTemplate->Parent, Level+1);
  return r;
}
//---------------------------------------------------------------------------
// Returns a string describing paragraph properties of StyleTemplate and
// its parents. Level must be 0. ParaStyleTemplate=True when describing a style
// template applied to a paragraph.
UnicodeString DescribeStyleTemplateText(TRVStyleTemplate* StyleTemplate, int Level,
  bool ParaStyleTemplate)
{
  if (!StyleTemplate)
    return "none";
    
  UnicodeString r = "\""+StyleTemplate->Name+"\"\r\n";
  UnicodeString s = DescribeTextProps(StyleTemplate->TextStyle, StyleTemplate->ValidTextProperties);
  if (!s.IsEmpty())
    r += "("+s+")";
  if (!StyleTemplate->Next && !ParaStyleTemplate)
    return r;
  if (!s.IsEmpty())
    r += ", ";
  if (Level==0)
    r += "based on ";
  else
    r += "which is based on ";
  if (! StyleTemplate->Parent)
    r += "default text attributes";
  else
    r += DescribeStyleTemplateText(StyleTemplate->Parent, Level+1, ParaStyleTemplate);
  return r;
}
//---------------------------------------------------------------------------
// Describing current text and paragraph formatting of rve.
// The description is shown in rvInfo
void TfrmMain::DescribeFormatting()
{
  rvInfo->Clear();
  if (!rve->ItemCount)
  {
    rvInfo->Format();
    return;
  }
  // paragraph attributes
  rvInfo->AddNL("Paragraph", 1, 0);
  TParaInfo* ParaInfo = rvs->ParaStyles->Items[rve->CurParaStyleNo];
  TRVStyleTemplateId Id = ParaInfo->StyleTemplateId;
  TRVStyleTemplate* ParaStyleTemplate = rvs->StyleTemplates->FindItemById(Id);
  ParaStyleTemplate->UpdateModifiedParaStyleProperties(ParaInfo);
  UnicodeString s = DescribeParaProps(ParaInfo, ParaInfo->ModifiedProperties);
  if (! s.IsEmpty())
  {
    rvInfo->AddNL("Inline attributes: ", 2, 0);
    rvInfo->Add(s, 0);
  }
  rvInfo->AddNL("Style: ", 2, 0);
  rvInfo->AddTextNL(DescribeStyleTemplatePara(ParaStyleTemplate, 0), 0, -1, 0);
  rvInfo->AddBreak();
  // text attributes
  rvInfo->AddNL("Text", 1, 0);
  TFontInfo* TextInfo = rvs->TextStyles->Items[rve->CurTextStyleNo];
  Id = TextInfo->StyleTemplateId;
  TRVStyleTemplate* StyleTemplate = rvs->StyleTemplates->FindItemById(Id);
  StyleTemplate->UpdateModifiedTextStyleProperties(TextInfo, ParaStyleTemplate);
  s = DescribeTextProps(TextInfo, TextInfo->ModifiedProperties);
  if (! s.IsEmpty())
  {
    rvInfo->AddNL("Inline attributes: ", 2, 0);
    rvInfo->Add(s, 0);
  }
  rvInfo->AddNL("Text style: ", 2, 0);
  rvInfo->AddTextNL(DescribeStyleTemplateText(StyleTemplate, 0, false), 0, -1, 0);
  rvInfo->AddNL("Paragraph style: ", 2, 0);
  rvInfo->AddTextNL(DescribeStyleTemplateText(ParaStyleTemplate, 0, true), 0, -1, 0);
  rvInfo->Format();
}
//---------------------------------------------------------------------------
void __fastcall TfrmMain::btnStylesClick(TObject *Sender)
{
  TfrmStyleList* frm = new TfrmStyleList(Application);
  frm->lst->Items->Assign(cmbStyles->Items);
  frm->lst->ItemIndex = cmbStyles->ItemIndex;
  frm->rvs->StyleTemplates->AssignStyleTemplates(rvs->StyleTemplates, true);
  if (frm->ShowModal()==mrOk)
    rve->ChangeStyleTemplates(frm->rvs->StyleTemplates);
  delete frm;
}
//---------------------------------------------------------------------------

