{-------------------------------------------------------------------------------
  Working with RVF and RTF files containing shared images.
  This demo stores them in the special subdirectory, but you can store them
  in a database, etc.
  How to save pictures as references in RVF:
    rvfoSavePicturesBody is EXCLUDED from RichViewEdit1.RVFOptions.
  How to save pictures as references in RTF:
    see RichViewEdit1SaveItemToFile.
  How to load pictures when reading such RVF files:
    see RichViewEdit1RVFPictureNeeded.
  How to load pictures when reading such RTF files:
    see RichViewEdit1ImportPicture.

  Paths to image file names are stored:
  - in rvespImageFileName properties for pictures
  - table.BackgroundImageFileName for tables
  - cell.BackgroundImageFileName for table cells.
  Normally, full paths to images are stored there.
  In this demo, path is stored relative to the application directory.
  Saved RTF and HTML files will be opened normally in other applications
  only if they are stored in the application directory.
-------------------------------------------------------------------------------}
{ Version 3.0
  Changes in v3.0:
  - compatibility with TRichView 14 (new parameters in OnRVFPictureNeeded
  - table and table cell background images are supported (in RVF)
  Changes in v2.0:
  - RTF support;
  - HTML saving;  
  - storing file names in rvespImageFileName property instead of item names;
  - storing paths relative to the application path instead of storing just
    file names (allows external applications to open RTF and HTML files
    correctly, if they are saved in the application path)
-------------------------------------------------------------------------------}

unit Unit1;

interface

uses
  AnsiStrings, Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, RVStyle, RVScroll, RichView, RVEdit, RVFuncs, StdCtrls,
  CRVData, RVTable, RVItem, RVTypes, RVGrHandler;

type
  TForm1 = class(TForm)
    RichViewEdit1: TRichViewEdit;
    RVStyle1: TRVStyle;
    Button1: TButton;
    OpenDialog1: TOpenDialog;
    Button2: TButton;
    Button3: TButton;
    OpenDialog2: TOpenDialog;
    SaveDialog1: TSaveDialog;
    procedure Button1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure RichViewEdit1Copy(Sender: TObject);
    procedure RichViewEdit1ImportPicture(Sender: TCustomRichView;
      const Location: String; Width, Height: Integer;
      var Graphic: TGraphic);
    procedure RichViewEdit1SaveItemToFile(Sender: TCustomRichView;
      const Path: String; RVData: TCustomRVData; ItemNo: Integer;
      SaveFormat: TRVSaveFormat; Unicode: Boolean; var OutStr: TRVRawByteString;
      var DoDefault: Boolean);
    procedure RichViewEdit1RVFPictureNeeded(Sender: TCustomRichView;
      const ItemName: String; Item: TRVNonTextItemInfo; Index1,
      Index2: Integer; var gr: TGraphic);
  private
    { Private declarations }
    function CopyImageToTheImagesDir(ImageFileName: String; gr: TGraphic): String;
    procedure SaveAllUnknownImages(RVData: TCustomRVData);
    procedure ConvertAllPathsToRelativePaths(RVData: TCustomRVData);
    function LoadGraphic(const FileName: String): TGraphic;    
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
  Randomize;
  OpenDialog2.InitialDir := ExtractFilePath(Application.ExeName);
  SaveDialog1.InitialDir := ExtractFilePath(Application.ExeName);
  RichViewEdit1.LoadRVF(ExtractFilePath(Application.ExeName)+'demo.rvf');
  RichViewEdit1.Format;
end;
{------------------------------------------------------------------------------}
// Returns graphic loaded from FileName.
// If loading fails, returns RVStyle1.InvalidPicture
function TForm1.LoadGraphic(const FileName: String): TGraphic;
begin
  Result := RVGraphicHandler.LoadFromFile(FileName);
  if Result=nil then begin
    Result := RVGraphicHandler.CreateGraphic(TGraphicClass(RVStyle1.InvalidPicture.Graphic.ClassType));
    Result.Assign(RVStyle1.InvalidPicture.Graphic);
  end;
end;
{------------------------------------------------------------------------------}
// RichViewEdit1.OnRVFPictureNeeded
// This event occurs when reading RVF files.
// This event loads this image from the Images subdirectory.
procedure TForm1.RichViewEdit1RVFPictureNeeded(Sender: TCustomRichView;
  const ItemName: String; Item: TRVNonTextItemInfo; Index1,
  Index2: Integer; var gr: TGraphic);
var FileName: String;
begin
  if Item is TRVGraphicItemInfo then
    FileName := TRVGraphicItemInfo(Item).ImageFileName
  else if Item is TRVTableItemInfo then
    if (Index1>=0) and (Index2>=0) then
      FileName := TRVTableItemInfo(Item).Cells[Index1, Index2].BackgroundImageFileName
    else
      FileName := TRVTableItemInfo(Item).BackgroundImageFileName
  else
    FileName := '';
  if FileName = '' then
    exit;
  FileName := ExtractFilePath(Application.ExeName)+FileName; // path is relative to the application path
  gr := LoadGraphic(FileName);
end;
{------------------------------------------------------------------------------}
// RichViewEdit1.OnImportPicture
// This event occurs when loading RTF files with external pictures
procedure TForm1.RichViewEdit1ImportPicture(Sender: TCustomRichView;
  const Location: String; Width, Height: Integer; var Graphic: TGraphic);
var FileName: String;
begin
  FileName := ExtractRelativePath(ExtractFilePath(Application.ExeName), Location);
  Graphic := LoadGraphic(FileName);
end;
{------------------------------------------------------------------------------}
// RichViewEdit1.OnSaveItemToFile
// Modifying how pictures are saved in RTF: saving as external pictures
procedure TForm1.RichViewEdit1SaveItemToFile(Sender: TCustomRichView;
  const Path: String; RVData: TCustomRVData; ItemNo: Integer;
  SaveFormat: TRVSaveFormat; Unicode: Boolean; var OutStr: TRVRawByteString;
  var DoDefault: Boolean);
var FileName: String;
begin
 if (SaveFormat=rvsfRTF) and
    ((RVData.GetItemStyle(ItemNo)=rvsPicture) or
     (RVData.GetItemStyle(ItemNo)=rvsHotPicture)) then begin
    RVData.GetItemExtraStrProperty(ItemNo, rvespImageFileName, FileName);
    OutStr := AnsiStrings.Format('{\field{\*\fldinst INCLUDEPICTURE "%s" \\d }}',
      [RVMakeRTFFileNameStr(FileName, RichViewEdit1.Style.DefCodePage,
       rvrtfDuplicateUnicode in RichViewEdit1.RTFOptions)]);
    DoDefault := False;
 end;
end;
{------------------------------------------------------------------------------}
// Inserting image.
// If this image is not from the Images subdirectory, copying it there
// (under the unique file name)
// Image file name is written in rvespImageFileName
//  (relative to the application path)
procedure TForm1.Button1Click(Sender: TObject);
var gr: TGraphic;
    ImageName: String;
begin
  if OpenDialog1.Execute then begin
    gr := RVGraphicHandler.LoadFromFile(OpenDialog1.FileName);
    if gr<>nil then begin
      // using relative path
      ImageName := ExtractRelativePath(ExtractFilePath(Application.ExeName),
        CopyImageToTheImagesDir(OpenDialog1.FileName, nil));
      RichViewEdit1.TopLevelEditor.BeginUndoGroup(rvutInsert);
      RichViewEdit1.TopLevelEditor.SetUndoGroupMode(True);
      try
        if RichViewEdit1.InsertPicture('', gr, rvvaBaseline) then
          RichViewEdit1.SetCurrentItemExtraStrProperty(rvespImageFileName, ImageName, True);
      finally
        RichViewEdit1.TopLevelEditor.SetUndoGroupMode(False);
      end;
      end
    else
      Application.MessageBox('Image loading error', 'Error', 0);
  end;
end;
{------------------------------------------------------------------------------}
// Copying the file ImageFileName to the images subdirectory (if gr=nil)
// or saving gr in the images subdirectory.
// Assigning an unique file name.
// Both ImageFileName and returned value are fully qualified paths.
function TForm1.CopyImageToTheImagesDir(ImageFileName: String; gr: TGraphic): String;
var ImagesDir, NewImageFileName, ImageExt: String;
    RandomValue: Integer;
begin
  ImageFileName := AnsiLowerCase(ImageFileName);
  ImagesDir := AnsiLowerCase(ExtractFilePath(Application.ExeName)+'Images\');
  if Pos(ImagesDir,ImageFileName)<>1 then begin
    NewImageFileName := ImagesDir+ExtractFileName(ImageFileName);
    if FileExists(NewImageFileName) then begin
      ImageExt := ExtractFileExt(NewImageFileName);
      NewImageFileName := Copy(NewImageFileName, 1, Length(NewImageFileName)-Length(ImageExt));
      RandomValue := Random(MaxInt);
      while FileExists(NewImageFileName+IntToStr(RandomValue)+ImageExt) do
        inc(RandomValue);
      NewImageFileName := NewImageFileName+IntToStr(RandomValue)+ImageExt;
    end;
    if gr=nil then
      CopyFile(PChar(ImageFileName), PChar(NewImageFileName), False)
    else
      gr.SaveToFile(NewImageFileName);
    Result := NewImageFileName;
    end
  else
    Result := ImageFileName;
end;
{------------------------------------------------------------------------------}
// Saving all images that not in the images directory
// Such images can appear when loading or pasting files with images
procedure TForm1.SaveAllUnknownImages(RVData: TCustomRVData);
var i,r,c: Integer;
    Tag: TRVTag;
    VAlign: TRVVAlign;
    table: TRVTableItemInfo;
    gr: TGraphic;
    s: TRVAnsiString;
    ImageFileName, Ext: String;
begin
  for i := 0 to RVData.ItemCount-1 do
    case RVData.GetItemStyle(i) of
      rvsPicture, rvsHotPicture:
        begin
          // pictures
          RVData.GetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
          ImageFileName := ExtractFilePath(Application.ExeName)+ImageFileName;
          if not (FileExists(ImageFileName)) then begin
            RVData.GetPictureInfo(i, s, gr, VAlign, Tag);
            Ext := GraphicExtension(TGraphicClass(gr.ClassType));
            ImageFileName := ExtractRelativePath(ExtractFilePath(Application.ExeName),
               CopyImageToTheImagesDir('Image.'+Ext, gr));
            RVData.SetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
          end;
        end;
      rvsTable:
        begin
          table := TRVTableItemInfo(RVData.GetItem(i));
          if table.BackgroundImage<>nil then begin
            // table background image
            ImageFileName := ExtractFilePath(Application.ExeName)+table.BackgroundImageFileName;
            if not (FileExists(ImageFileName)) then begin
              gr := table.BackgroundImage;
              Ext := GraphicExtension(TGraphicClass(gr.ClassType));
              ImageFileName := ExtractRelativePath(ExtractFilePath(Application.ExeName),
                 CopyImageToTheImagesDir('Image.'+Ext, gr));
              table.BackgroundImageFileName := ImageFileName;
            end;
          end;
          for r := 0 to table.Rows.Count-1 do
            for c := 0 to table.Rows[r].Count-1 do
              if table.Cells[r,c]<>nil then begin
                if table.Cells[r,c].BackgroundImage<>nil then begin
                  // cell background image
                  ImageFileName := ExtractFilePath(Application.ExeName)+table.Cells[r,c].BackgroundImageFileName;
                 if not (FileExists(ImageFileName)) then begin
                   gr := table.Cells[r,c].BackgroundImage;
                   Ext := GraphicExtension(TGraphicClass(gr.ClassType));
                   ImageFileName := ExtractRelativePath(ExtractFilePath(Application.ExeName),
                     CopyImageToTheImagesDir('Image.'+Ext, gr));
                   table.Cells[r,c].BackgroundImageFileName := ImageFileName;
                  end;
                end;
                // recursive call for cells
                SaveAllUnknownImages(table.Cells[r,c].GetRVData);
              end;
        end;
    end;
end;
{------------------------------------------------------------------------------}
// After loading from RTF, full file names are assigned to rvespImageFileName
// properties. This procedure is called to convert them to relative paths.
procedure TForm1.ConvertAllPathsToRelativePaths(RVData: TCustomRVData);
var i,r,c: Integer;
    table: TRVTableItemInfo;
    ImageFileName: String;
begin
  for i := 0 to RVData.ItemCount-1 do
    case RVData.GetItemStyle(i) of
      rvsPicture, rvsHotPicture:
        begin
          RVData.GetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
          ImageFileName := ExtractRelativePath(ExtractFilePath(Application.ExeName),
            ImageFileName);
          RVData.SetItemExtraStrProperty(i, rvespImageFileName, ImageFileName);
        end;
      rvsTable:
        begin
          table := TRVTableItemInfo(RVData.GetItem(i));
          for r := 0 to table.Rows.Count-1 do
            for c := 0 to table.Rows[r].Count-1 do
              if table.Cells[r,c]<>nil then
                ConvertAllPathsToRelativePaths(table.Cells[r,c].GetRVData);
        end;
    end;
end;
{------------------------------------------------------------------------------}
// Before copying to the clipboard
procedure TForm1.RichViewEdit1Copy(Sender: TObject);
begin
  SaveAllUnknownImages(RichViewEdit1.RVData);
end;
{------------------------------------------------------------------------------}
// Loading doc
procedure TForm1.Button3Click(Sender: TObject);
var r: Boolean;
begin
  if OpenDialog2.Execute then begin
    RichViewEdit1.Clear;
    r := False;
    case OpenDialog2.FilterIndex of
      1:  r := RichViewEdit1.LoadRVF(OpenDialog2.FileName);
      2:  begin
            r := RichViewEdit1.LoadRTF(OpenDialog2.FileName);
            ConvertAllPathsToRelativePaths(RichViewEdit1.RVData);
          end;
    end;
    RichViewEdit1.Format;
    if not r then
      Application.MessageBox('Document loading error', 'Error', 0);
  end;
end;
{------------------------------------------------------------------------------}
// Saving doc
procedure TForm1.Button2Click(Sender: TObject);
var r: Boolean;
begin
  if SaveDialog1.Execute then begin
    SaveAllUnknownImages(RichViewEdit1.RVData);
    r := False;
    case SaveDialog1.FilterIndex of
      1: r := RichViewEdit1.SaveRVF(SaveDialog1.FileName, False);
      2: r := RichViewEdit1.SaveRTF(SaveDialog1.FileName, False);
      3: r := RichViewEdit1.SaveHTMLEx(SaveDialog1.FileName, 'Shared Image Demo',
        'img', '', '', '', [rvsoUseCheckpointsNames, rvsoUseItemImageFileNames]);
    end;
    if not r then
      Application.MessageBox('Document saving error', 'Error', 0);
  end;
end;



end.
